local building = {}
building.__index = building

local verticalSlices,horizontalSlices
verticalSlices = {}
horizontalSlices = {}
for i=0,30 do
  table.insert(verticalSlices, love.graphics.newQuad(31-i,0,2,160,32,160))
  table.insert(horizontalSlices, love.graphics.newQuad(0,31-i,160,2,160,32))
end

local function new(x,y,width,height,floors,leftSide,rightSide,front,roof)
  return setmetatable({
    x = x,
    y = y,
    color = math.random(150,200),
    width = width,
    height = height,
    floors = floors,
    leftImage = leftSide,
    rightImage = rightSide,
    frontImage = front,
    prepared = false,
    roofQuad = love.graphics.newQuad(0,0,width,height,16,16),
    roofCoords = {0,0},
    roofImage = roof
  },building)
end

function building:update(playerX,playerY,distance)
  if not self.prepared then return end
  -- local distance = 150
  local edges = {}
  local growth = 0.005
  local topLeft,topRight,bottomLeft,bottomRight = {},{},{},{}
  for i=0,30*self.floors do
    topLeft[i+1] = {
      px = math.floor((distance * (self.x - playerX)) / (distance + -i*growth*distance) + playerX),
      py = math.floor((distance * (self.y - playerY)) / (distance + -i*growth*distance) + playerY),
    }
    topRight[i+1] = {
      px = math.floor((distance * (self.x + self.width - playerX)) / (distance + -i*growth*distance) + playerX),
      py = math.floor((distance * (self.y - playerY)) / (distance + -i*growth*distance) + playerY),
    }
    bottomRight[i+1] = {
      px = math.floor((distance * (self.x + self.width - playerX)) / (distance  + -i*growth*distance) + playerX),
      py = math.floor((distance * (self.y + self.height - playerY)) / (distance + -i*growth*distance) + playerY),
    }
    bottomLeft[i+1] = {
      px = math.floor((distance * (self.x - playerX)) / (distance  + -i*growth*distance) + playerX),
      py = math.floor((distance * (self.y + self.height - playerY)) / (distance + -i*growth*distance) + playerY),
    }
  end
  local finalTL, finalBr = topLeft[(self.floors*30)+1], bottomRight[(self.floors*30)+1]
  self.roofQuad:setViewport(0,0, math.abs(finalTL.px - finalBr.px), math.abs(finalTL.py - finalBr.py))
  self.roofCoords = {finalTL.px+1,finalTL.py+1}

  self:updateBatch(self.leftBatch,verticalSlices,topLeft,bottomLeft)
  self:updateBatch(self.rightBatch,verticalSlices,topRight,bottomRight)
  self:updateBatch(self.backBatch,horizontalSlices,topLeft,topRight)
  self:updateBatch(self.frontBatch,horizontalSlices,bottomLeft,bottomRight)
end

function building:prepare(left,right,front,back)
  self.leftBatch = left
  self.rightBatch = right
  self.frontBatch =  front
  self.backBatch = back
  self.prepared = true
end

function building:unprepare()
  if not prepared then return end
  self.leftBatch = nil
  self.rightBatch = nil
  self.frontBatch =  nil
  self.backBatch = nil
  self.prepared = true
end

function building:updateBatch(batch,slices,drawCoords,scaleCoords)
  batch:clear()
  for i,drawCoord in ipairs(drawCoords) do
    local scale = math.max(math.abs(drawCoord.py - scaleCoords[i].py),math.abs(drawCoord.px - scaleCoords[i].px))/160
    batch:addq(slices[(i%31)+1], drawCoord.px, drawCoord.py,0,scale)
  end
end

function building:draw(playerX,playerY)
  if not self.prepared then return end
  love.graphics.setColor(self.color,self.color,self.color)
  love.graphics.draw(self.backBatch)
  if playerY < self.y + self.height then
    love.graphics.draw(self.frontBatch)
    if playerY < self.y then
      if playerX > self.x + self.width then
        love.graphics.draw(self.leftBatch)
        love.graphics.draw(self.rightBatch)
      else
        love.graphics.draw(self.rightBatch)
        love.graphics.draw(self.leftBatch)
      end
      love.graphics.draw(self.backBatch)
    else
      love.graphics.draw(self.backBatch)
      if playerX > self.x + self.width then
        love.graphics.draw(self.leftBatch)
        love.graphics.draw(self.rightBatch)
      else
        love.graphics.draw(self.rightBatch)
        love.graphics.draw(self.leftBatch)
      end
    end
  else
    love.graphics.draw(self.backBatch)
    if playerX > self.x + self.width then
      love.graphics.draw(self.leftBatch)
      love.graphics.draw(self.rightBatch)
    else
      love.graphics.draw(self.rightBatch)
      love.graphics.draw(self.leftBatch)
    end
    love.graphics.draw(self.frontBatch)
  end
  love.graphics.setColor(180,180,180)
  if self.floors < 6 then
    love.graphics.drawq(self.roofImage, self.roofQuad, unpack(self.roofCoords))
  end
  
  -- self:unload()
end

-- the module
return setmetatable({new = new, isvector = isvector},
{__call = function(_, ...) return new(...) end})