local farts = {}
farts.__index = farts

local function new(fartImages)
  return setmetatable({
      farts = {},
      fartImages = fartImages
    }, farts)
end

function farts:create(world, player)
  local newFart = {}
  local vx,vy = player.body:getLinearVelocity()
  local magnitude = math.sqrt(vx * vx + vy * vy)
  local dx, dy = math.sin(player.angle), math.cos(player.angle)
  local ax,ay
  -- love.graphics.setCaption(#powerups)
  if vx == 0 then
    ax = 0
  else
    ax = vx/math.abs(vx)
  end
  if vy == 0 then
    ay = 0
  else
    ay = vy/math.abs(vy)
  end

  local x, y = player.body:getWorldCenter()
  if lastX then
    x = lastX
  end
  newFart.growthRate = player.fartiness
  newFart.body = love.physics.newBody(world, x+(3*ax), y+(5*ay), "dynamic") --place the body in the center of the world and make it dynamic, so it can move around
  newFart.shape = love.physics.newCircleShape(2) --the player's shape has a radius of 20
  newFart.fixture = love.physics.newFixture(newFart.body, newFart.shape, 1) -- Attach fixture to body and give it a density of 1.
  newFart.fixture:setCategory(2)
  newFart.fixture:setMask(2,6)
  newFart.body:setMass(3*newFart.growthRate)
  newFart.particles = love.graphics.newParticleSystem(fartImages[1], 20*newFart.growthRate)
  newFart.particles:setEmissionRate          (30)
  newFart.particles:setLifetime              (1)
  newFart.particles:setParticleLife          (1,3)
  newFart.particles:setPosition              (0, 0)
  newFart.particles:setDirection             (0)
  newFart.particles:setSpread                (600)
  newFart.particles:setSpeed                 (10, 20, 100*player.speed)
  newFart.particles:setGravity               (0)
  newFart.particles:setRadialAcceleration    (1, 50*newFart.growthRate)
  newFart.particles:setTangentialAcceleration(2)
  newFart.particles:setSizes                 (1,2,3,5*newFart.growthRate)
  newFart.particles:setSizeVariation         (1)
  newFart.particles:setRotation              (0.3)
  newFart.particles:setSpin                  (1)
  newFart.particles:setSpinVariation         (10)
  newFart.particles:setColors                (255,255,255,80, 155,155,155,80, 85,85,85,80)
  local vSpeed
  if not gameOver then
    if ay < 0 then
      vSpeed = vy*65
    else
      vSpeed = vy*65
    end
    newFart.body:applyForce(dx*magnitude*100 + math.random(-100,100),ay*math.abs(dy)*math.max(newFart.growthRate*3,1)*magnitude*65)
  else
    newFart.body:setY(20 + 10*newFart.growthRate)
    newFart.body:applyLinearImpulse(150*newFart.growthRate*math.sin(love.timer.getMicroTime()*50),0)
  end
  newFart.ttl = math.random(300,400)
  table.insert(self.farts, newFart)
  newFart = nil
end

function farts:update(dt)
  for i,fart in ipairs(self.farts) do
    if not fart then 
      break
    end
    if fart.ttl < 0 then
      fart.particles:stop()
      fart.particles = nil
      fart.fixture:destroy()
      table.remove(self.farts, i)
    else
      if fart.ttl < 80 then
        fart.particles:setSprite(fartImages[3])
      elseif fart.ttl < 300 then
        fart.particles:setSprite(fartImages[2])
      end
      fart.ttl = fart.ttl - (dt*100)
      fart.particles:update(dt)
      local r = fart.shape:getRadius() + fart.growthRate
      fart.shape:setRadius(r)
      fart.fixture:getShape():setRadius(r)
    end
  end
end

function farts:draw()
  for i,fart in ipairs(self.farts) do
    if fart.body then
      love.graphics.setColor(255,255,255, math.min(fart.ttl,20))
      love.graphics.draw(fart.particles, fart.body:getX(), fart.body:getY())
    end
  end
  love.graphics.setColor(255,255,255,255)
end

-- the module
return setmetatable({new = new},
  {__call = function(_, ...) return new(...) end})