require('anal')
local Camera = require('camera')
local Player = require('player')
local FartManager = require('farts')
local PedestrianManager = require('pedestrians')
Building = require('building')

local game = {}
local leftBatches,rightBatches,backBatches,frontBatches = {},{},{},{}
local buildings = {}
local drawableBuildings


local function boundBox(x,y,width,height)
  local body,shape,fixture
  body = love.physics.newBody(world, x+((width)/2) ,y+((height)/2))
  shape = love.physics.newRectangleShape(width,height)
  fixture = love.physics.newFixture(body,shape)
  fixture:setCategory(5)
end

local function buildBuilding(x,y,width,height,floors)
  local newBuilding = Building.new(x*320,y*320,width,height,floors,leftImage,rightImage,bottomImage,roofImage)
  if not buildings[x] then
    buildings[x] = {}
  end
  buildings[x][y] = newBuilding
  boundBox(x*320,y*320,width,height)
end

local function updateDrawable()
  drawableBuildings = {} -- reinitialize the array so we can rebuild it
  local cx,cy
  cx,cy = cam:worldCoords(0,0)
  startX = math.floor(cx/320)
  startY = math.floor(cy/320)
  
  for x=startX,startX+3 do
    if buildings[x] then
      for y=startY,startY+3 do

        if buildings[x][y] then
          local i,building = #drawableBuildings + 1,buildings[x][y]
          table.insert(drawableBuildings, building)    
        end
      end
    end    
  end
end

function drawBuildings()
  if not drawableBuildings then return end
  table.sort(drawableBuildings,function(a,b)
    local dxA,dyA,dxB,dyB,dtx,dmx,dbx,dty,dmy,dby, distA, distB
    dtx = math.abs(a.x - playerX)
    dbx = math.abs(a.x + (a.width) - playerX)
    dty = math.abs(a.y - playerY)
    dby = math.abs(a.y + (a.height) - playerY)
    dxA = math.min(dtx,dbx)
    dyA = math.min(dty,dby)
    distA = math.sqrt(dxA*dxA + dyA*dyA)

    dtx = math.abs(b.x - playerX)
    dbx = math.abs(b.x + (b.width) - playerX)
    dty = math.abs(b.y - playerY)
    dby = math.abs(b.y + (b.height) - playerY)
    dxB = math.min(dtx,dbx)
    dyB = math.min(dty,dby)
    distB = math.sqrt(dxB*dxB + dyB*dyB)
    return distA > distB
  end)
  for i,building in ipairs(drawableBuildings) do
    -- building:update(playerX,playerY)
    building:draw(player.body:getX(),player.body:getY())
  end
end

local effects = {
    taco = function() 
      player.angle = -player.angle + (math.pi*love.timer.getTime())--math.sin(love.timer.getMicroTime()*20)*0.3 -- wavy farts
    end,
  
    hotdog = function() 
      player.fartiness = math.min(player.fartiness + 1,2.3) -- big farts
    end,

    beer = function()
      player.scale = math.min(player.scale + 0.3, 3.5)
      player.body:setMass(math.min(player.body:getMass()+5),25) -- some kind of fart thing out the back 
      player.speed = player.body:getMass()*1000
    end,

    curry = function()
      player.remainingFarts = player.maxFarts -- infinite farts
    end,

    medicine = function()
      player.remainingFarts = math.min(player.maxFarts/2, player.remainingFarts) -- gas-b-gone
    end
    
  }

local function applyEffects(dt)
  player:renorm()
  
  local effect
  for i=1,#player.activeEffects do
    if player.activeEffects[i] then
      effect = player.activeEffects[i]
      effect.ttl = effect.ttl - dt
      effects[effect.kind]() -- apply
      if effect.ttl < 0 then
        table.remove(player.activeEffects, i)
      end
    end
  end
end

local function reactToFart(a,b,contact)
  local person
  if a:getCategory() == 1 then
    person = a
  elseif b:getCategory() == 2 then
    person = b
  end

  if not person then return end
  local vx, vy = contact:getNormal()

  person:getBody():applyLinearImpulse(-vx,-vy)
  local idx = person:getUserData().idx
  if not disgustedPatrons[idx] then disgustCount = disgustCount + 1 end
  if (disgustedPatrons[idx] or 0) > 0 then return end -- already said something
  
  local i = math.random(1,#disgustSounds)
  local sound = disgustSounds[i]
  disgustedPatrons[idx] = math.random(1,4)/5
  if sound:isStopped() then
    disgustSounds[i]:setPitch(math.random(7,10)/10)
    disgustSounds[i]:play()
    return
  end
end

local function collideWithPatron()
  collisionCount = collisionCount+1
end

local function consumePowerup(a,b,contact)
  local powerup
  if a:getCategory() == 7 then
    powerup = a
  elseif b:getCategory() == 7 then
    powerup = b
  else
    return
  end

  player.remainingFarts = player.maxFarts
  contact:setEnabled(false)
  table.insert(player.activeEffects, {kind =  powerup:getUserData().kind, tts = math.random(100,175)/1000, ttl = 5})
  powerups[powerup:getUserData().idx] = nil
  powerup:destroy()
  powerup = nil
end

local function collide(a,b,contact) -- 1: pedestrian, 2: fart, 3: player, 7: powerup, 5: wall, 9: win
  if gameOver then return end
  local collisionType = a:getCategory() + b:getCategory()
  if collisionType == 3 then -- 1 and 2
    reactToFart(a,b,contact)
  elseif collisionType == 4 then -- 3 and 1
    collideWithPatron()
  elseif collisionType == 10 then -- 3 and 7
    consumePowerup(a,b,contact)
  elseif collisionType == 12 then -- 3 and 9
    gameOver = true
    lastX = player.body:getX()
  end
  collisionType = nil
end

local function placePowerup(x,y,kind)
  local kinds = {'hotdog', 'curry', 'beer', 'taco', 'medicine'}
  local newPowerup,extra = {},{}

  
  newPowerup.body = love.physics.newBody(world, x, y, "kinematic") --place the body in the center of the world and make it dynamic, so it can move around
  newPowerup.shape = love.physics.newCircleShape(20) --the player's shape has a radius of 20
  newPowerup.fixture = love.physics.newFixture(newPowerup.body, newPowerup.shape, 0) -- Attach fixture to body and give it a density of 1.
  newPowerup.fixture:setCategory(7)
  newPowerup.fixture:setMask(1,2)
  extra.kind = kinds[kind+1]
  newPowerup.image = powerupImages[extra.kind]
  extra.idx = #powerups + 1
  newPowerup.fixture:setUserData(extra)
  -- if newPowerup.kind then
  powerups[extra.idx] = newPowerup
  -- end
  newPowerup = nil
  extra = nil
end

local function initializePedestrians(first,last)
  for y=first,last do
    if pedestrian_locations[y] then
      for i,data in ipairs(pedestrian_locations[y]) do
        pedestrians:add(world,(data[1] * 32) + math.random(0,10)-5,(y*32) + math.random(0,10)-5, walkCycles[math.random(3)], data[2])
      end
      pedestrian_locations[y] = nil
    end
  end
end

local function makeBounds(width,height)
  local topLeft,topRight,bottomLeft,bottomRight,horizontal,vertical
  topLeft = {0,32}
  topRight = {width*32,0}
  bottomLeft = {0, height*32}
  bottomRight = {width*32, height*32}

  horizontal = {0,0,width*32,0}
  vertical =  {0,0,0,height*32}

  local body = love.physics.newBody(world, unpack(topLeft))
  local shape = love.physics.newEdgeShape(unpack(horizontal))
  local fixture = love.physics.newFixture(body,shape)
  fixture:setCategory(9)

  body = love.physics.newBody(world, unpack(topLeft))
  shape = love.physics.newEdgeShape(unpack(vertical))
  fixture = love.physics.newFixture(body,shape)
  fixture:setCategory(5)

  body = love.physics.newBody(world, unpack(bottomLeft))
  shape = love.physics.newEdgeShape(unpack(horizontal))
  fixture = love.physics.newFixture(body,shape)
  fixture:setCategory(5)

  body = love.physics.newBody(world, unpack(topRight))
  shape = love.physics.newEdgeShape(unpack(vertical))
  fixture = love.physics.newFixture(body,shape)
  fixture:setCategory(5)
  topLeft,topRight,bottomLeft,bottomRight,horizontal,vertical = nil,nil,nil
end

local function drawWalls()
  love.graphics.draw(streetBatch)
  love.graphics.setColor(255,255,255)
  local leftScale,rightScale,leftX, rightX
  leftScale = math.min(math.max(player.body:getX()/600,0.1),1.3)
  rightScale = math.min(math.max((map:getWidth()*32 - player.body:getX())/600,0.1),1.3)
  leftX = -leftScale*64
  rightX = map:getWidth()*32
  love.graphics.draw(bathrooms, -32, -110)
  love.graphics.drawq(streetLeft, buildingsQuad, leftX, -128, 0, leftScale, 4, 0, 0, 0, 0.5)
  love.graphics.drawq(streetRight, buildingsQuad, rightX, 0, 0, rightScale, 4, 0, 0, 0, -0.5)
  love.graphics.drawq(streetRoofLeft, roofQuad, leftX - 256, -128, 0, 1, 4)
  love.graphics.drawq(streetRoofRight, roofQuad, rightX + (64*rightScale), -128, 0, 1, 4)
  
end

local function draw_goal()
  -- local x = love.player
  -- love.graphics.draw
end

local function draw_hud()
  if gamePaused then
    love.graphics.print("press p to unpause", half_width - 160,half_height)
  end

  local fartsLeft = player.remainingFarts/player.maxFarts
  scoreShader:send('fartsLeft', fartsLeft)
  -- local 
  local alpha = fartsLeft *255

  local elapsed = currentTime - startTime
  local minutes = math.floor(elapsed/60)
  local seconds = elapsed % 60
  local msec = (seconds % 1) * 1000

  -- love.graphics.print(love.timer.getFPS().."fps   "..string.format("% i",math.floor(fartsLeft*100)).."%  "..minutes..string.format("%0i:%02d.%03d", minutes, seconds,msec).." "..disgustCount..", "..collisionCount,5,5)
  love.graphics.setColor(50,255,50)
  love.graphics.print(string.format("%02i:%02d.%03d", minutes, seconds,msec),5,5)
  love.graphics.setColor(255,255,255)
  love.graphics.draw(lower_gi, world_width - 58, 58, 0, 1, 1,48,48)
  love.graphics.setPixelEffect(scoreShader)
  love.graphics.draw(score_mask, world_width - 58, 58, 0, 1, 1,48,48)
  love.graphics.setPixelEffect()
  -- love.graphics.draw(maskedScore, world_width - 58, 58, 0, 1, 1,48,48)
  
  love.graphics.setColor(50,255,50,alpha)
  love.graphics.print('z', world_width - 90, 80)
  love.graphics.setColor(255,255,255)
  if countingDown then
    love.graphics.scale(3)
    if math.ceil(countingDown) > 0 then
      love.graphics.print(math.ceil(countingDown), 126, 16)
    else
      love.graphics.print('go!',112,16)
    end
  end

  -- draw_goal()

  if gameOver then
    love.graphics.print("pedestrians disgusted: "..disgustCount, 64, 400)
    love.graphics.print("pedestrians bumped: "..collisionCount, 64, 430)
    love.graphics.print("final time: "..string.format("%02i:%02d.%03d", minutes, seconds,msec), 64, 460)
    love.graphics.print('r to reset', 64, 500)
    love.graphics.print('esc for menu', 64, 540)
    love.graphics.scale(3)
    local sinTime,cosTime = math.abs(math.sin(love.timer.getMicroTime())), math.abs(math.cos(love.timer.getMicroTime()))
    love.graphics.setColor(155*sinTime,255*cosTime,155*sinTime)
    love.graphics.print('you made it!', 20, 20)
    love.graphics.setColor(255,255,255)
  end
  -- maskedScore, score_mask = nil, nil
end

local function drawEntities()
  pedestrians:draw()
  love.graphics.setColor(255,255,255)
  farts:draw()

  for i,powerup in ipairs(powerups) do
    if powerup then
      love.graphics.draw(powerup.image, powerup.body:getX(), powerup.body:getY() + math.sin(love.timer.getMicroTime()*15)*5, 0, 2, 2, 8,8)
    end
  end
  drawBuildings()
  if gameOver then return end
  player:draw()
end

local function draw_world()
  if gamePaused then 
    return 
  end
  drawWalls()
  drawEntities()
end

function game:init()
  
  love.physics.setMeter(64) --the height of a meter our worlds will be 64px

  score_mask = love.graphics.newImage('gfx/life_bar_mask.gif')
  scoreShader = love.graphics.newPixelEffect([[
    extern number fartsLeft;
    vec4 effect(vec4 color, Image texture, vec2 texture_coords, vec2 pixel_coords) {
      vec4 textureColor = Texel(texture, texture_coords);
      if (textureColor.w > 0 && textureColor.x <= fartsLeft) {
        return vec4(0,fartsLeft,0,0.8);
      } else {
        return vec4(0,0,0,0);
      }
    }
  ]])
  scoreShader:send('fartsLeft', 1.0)
  world_width = 800
  world_height = 600
  half_width = world_width/2
  half_height = world_height/2
  world = love.physics.newWorld(0,0,world_width,world_height,true)
  objects = {} -- table to hold all our physical objects
  world:setCallbacks(collide)

  disgustSounds = {}
  for i=1,6 do
    local disgust1 = love.audio.newSource('audio/disgust.wav', 'static')
    table.insert(disgustSounds, disgust1)
  end
  

  fartSound = love.audio.newSource('audio/fart_3.wav', "static")
  local playerX,playerY

  local img = love.graphics.newImage('gfx/player_body.gif')
  playerBody = newAnimation(img, 64, 64, 0.1, 6)
  local img = love.graphics.newImage('gfx/player_legs.gif')
  playerLegs = newAnimation(img, 64, 64, 0.1, 6)
  curbL = love.graphics.newImage('gfx/curb_l.gif')
  curbR = love.graphics.newImage('gfx/curb_r.gif')
  street = love.graphics.newImage('gfx/street.gif')
  sidewalk = love.graphics.newImage('gfx/sidewalk.gif')

  curbL:setWrap('repeat', 'repeat')
  curbR:setWrap('repeat', 'repeat')
  street:setWrap('repeat', 'repeat')
  sidewalk:setWrap('repeat', 'repeat')

  fartFrame = love.graphics.newImage('gfx/fart_frame.gif')
  fartImages = { love.graphics.newImage('gfx/farticles1.png'), love.graphics.newImage('gfx/farticles2.png'), love.graphics.newImage('gfx/farticles3.png')}
  lower_gi = love.graphics.newImage('gfx/lower_gi.png')
  streetTiles = love.graphics.newImage('gfx/simple_street_tiles.png')
  
  roofImage = love.graphics.newImage('gfx/roof.png')
  roofImage:setWrap('repeat','repeat')
  leftImage = love.graphics.newImage('gfx/left.png')
  leftImage:setWrap('repeat','repeat')
  leftImage:setFilter('nearest','nearest')
  rightImage = love.graphics.newImage('gfx/right.png')
  rightImage:setWrap('repeat','repeat')
  rightImage:setFilter('nearest','nearest')
  bottomImage = love.graphics.newImage('gfx/bottom.png')
  bottomImage:setFilter('nearest','nearest')

  for i=1,17 do -- make a pool of spritebatches for on-screen buildings
    leftBatches[i] =love.graphics.newSpriteBatch(leftImage, 32*6)
    rightBatches[i] = love.graphics.newSpriteBatch(rightImage, 32*6)
    frontBatches[i] = love.graphics.newSpriteBatch(bottomImage, 32*6)
    backBatches[i] = love.graphics.newSpriteBatch(bottomImage, 32*6)
  end

  powerupImages = {
    hotdog = love.graphics.newImage('gfx/food_hotdog.png'),
    taco = love.graphics.newImage('gfx/food_taco.png'),
    curry = love.graphics.newImage('gfx/food_curry.png'),
    beer = love.graphics.newImage('gfx/food_beer.png'),
    medicine = love.graphics.newImage('gfx/food_gas_no.png')
  }

  map = love.image.newImageData('gfx/map_1.bmp')
  tiles_map = love.image.newImageData('gfx/tiles_map.bmp')

  streetBatch = love.graphics.newSpriteBatch(streetTiles, map:getWidth() * map:getHeight() * 32)
  mapQuads = {}
  for i=0,5 do
    mapQuads[i] = love.graphics.newQuad(i*16, 0, 16, 16, 96, 16)
  end

  walkCycles = {love.graphics.newImage('gfx/walkcycle_1.png'),love.graphics.newImage('gfx/walkcycle_2.png'),love.graphics.newImage('gfx/walkcycle_3.png')}

  portoPotty = love.graphics.newImage('gfx/portopotty.gif')
  streetLeft = love.graphics.newImage('gfx/longstreet.gif')
  streetRight = love.graphics.newImage('gfx/longstreetR.gif')
  streetRoofLeft = love.graphics.newImage('gfx/longstreetRoof.gif')
  streetRoofRight = love.graphics.newImage('gfx/longstreetRRoof.gif')
  bathrooms = love.graphics.newImage('gfx/bathrooms.png')
  streetRight:setWrap('clamp', 'repeat')
  streetLeft:setWrap('clamp', 'repeat')
  streetRoofRight:setWrap('clamp', 'repeat')
  streetRoofLeft:setWrap('clamp', 'repeat')
  buildingsQuad = love.graphics.newQuad(0,0,64,map:getHeight()*32,64,1024)
  roofQuad = love.graphics.newQuad(0,0,256,map:getHeight()*32,3,1024)
  worldBounds = {0,0,map:getWidth()*32,0,map:getWidth()*32,map:getHeight()*32,0,map:getHeight()*32}
  -- sidewalkQuad = love.graphics.newQuad(0,0,64,map:getHeight()*32,64,64)
  -- streetQuad = love.graphics.newQuad(0,0,(map:getWidth()*32) - 256, map:getHeight()*32, 64,64)
  makeBounds(map:getWidth(), map:getHeight())
  cam = Camera(0,0)
    --initial graphics setup
  love.graphics.setBackgroundColor(0,0,0) --set the background color to a nice blue
end

function game:enter(state)
  if world:getBodyCount() > 0 then
    local bodyList = world:getBodyList()
    for i,body in ipairs(bodyList) do
      for i,fixture in pairs(body:getFixtureList()) do
        fixture:setUserData(nil)
        fixture:destroy()
      end
      body = nil
    end
    bodyList = nil
  end
  collectgarbage()
  makeBounds(map:getWidth(),map:getHeight())
  streetBatch:clear()
  zoomLevel = 1

  startTime = love.timer.getTime()
  currentTime = startTime
  countingDown = 3
  gamePaused = false
  gameOver = false
  lastX = nil
  disgustCount = 0
  collisionCount = 0

  powerups = nil
  powerups = {}
  
  pedestrian_locations = nil
  pedestrian_locations = {}


  if pedestrians then

    pedestrians.pedestrians = nil
    pedestrians.drawable = nil
    pedestrians = nil
  end
  pedestrians = PedestrianManager.new()

  local r,g,b,a
  for x = 0, map:getWidth() - 1 do
    for y = 0, map:getHeight() - 1 do
      r,g,b,a = map:getPixel(x,y) -- Entities
      if r == 255 and b == 0 then
        if pedestrian_locations[y] then
          table.insert(pedestrian_locations[y], {x,g})
        else
          pedestrian_locations[y] = {{x,g}}
        end
      elseif b == 255 and r + g == 0 then
        playerX,playerY = x*32,y*32
      elseif b == 255 and r == 255 and g < 255 then
        placePowerup(x*32,y*32, g)
      elseif r + b == 0 then
        placeStaticTile(x*32,y*32, g)
      end

      r,g,b,a = tiles_map:getPixel(x,y)
      if r + g + b == 0 then
        streetBatch:addq(mapQuads[0],x*32,y*32,0,2,2) 
      elseif r == 255 and g + b < 100 then
        streetBatch:addq(mapQuads[2],x*32,y*32,0,2,2)
      elseif r == 255 and g + b > 100 then
        streetBatch:addq(mapQuads[3],x*32,y*32,0,2,2)
      elseif r + b > 450 then
        streetBatch:addq(mapQuads[4],x*32,y*32,0,2,2)
      elseif b == 255 then
        streetBatch:addq(mapQuads[5],x*32,y*32,0,2,2)
      else
        streetBatch:addq(mapQuads[1],x*32,y*32,0,2,2) 
      end
    end
  end


  if player then
    player = nil
  end

  buildings = {}
  
  for y=1,100 do
   buildBuilding(1,y,160,160,2)
  end
  

  player = Player.new(world, playerX, playerY, playerLegs, playerBody, fartFrame)

  cam = cam:lookAt(playerX, playerY)
  
  lowestY = playerY + world_height
  
  if disgustedPatrons then
    disgustedPatrons = nil
  end
  disgustedPatrons = {}
  if farts then
    farts.fartImages = nil
    farts.farts = nil
    farts = nil
  end
  farts = FartManager.new(fartImages)
  collectgarbage('collect')
end

function game:keypressed(key)
  if key == 'p' then
    gamePaused = not gamePaused
  end
  if key == 'r' then
    collectgarbage('collect')
    game:enter()
  end
  if key == 'escape' then
    game:quit()
  elseif key == 'f' then
    love.graphics.toggleFullscreen()
  elseif key == '1' then
    table.insert(player.activeEffects, {kind='hotdog', ttl=2})
  elseif key == '2' then
    table.insert(player.activeEffects, {kind='beer', ttl=2})
  elseif key == '3' then
    table.insert(player.activeEffects, {kind='curry', ttl=2})
  end
end

function game:keyreleased(key)
  if key == "z" then
    fadeFart = true
  end
end

local function countDown(dt)
  if math.floor(countingDown) == 0 then
    startTime = love.timer.getTime()
    currentTime = startTime
  end
  countingDown = countingDown - dt
end

function game:update(dt)
  if gamePaused then 
    return 
  end
  dt = math.min(dt, 1000/60)
  local fartsLeft = player.remainingFarts/player.maxFarts
  local vx,vy = player.body:getLinearVelocity()
  local requestedZoom
  if math.abs(vy) > 0 then
    requestedZoom = math.min(math.max(math.min(math.max(250/math.abs(vy),0.8),1) - zoomLevel,-0.002),0.001)
    cam:zoomTo(zoomLevel + requestedZoom)
  end

  zoomLevel = cam.scale
  
  
  cam:lookAt(math.min(math.max(player.body:getX(),300*cam.scale),650*cam.scale), math.min(player.body:getY(),lowestY - half_height))
  local y = math.floor(player.body:getY()/32)
  initializePedestrians(math.max(y-20,0),y+20)
  

  table.foreach(disgustedPatrons, function(idx,timer)
    if timer <= 0 then 
      disgustedPatrons[idx] = 0
      return 
    end
    timer = timer - dt
    
    disgustedPatrons[idx] = timer
  end)

  if fadeFart then
    local volume = fartSound:getVolume() - 0.1
    fartSound:setVolume(volume)
    if volume < 0.1 then
      fartSound:stop()
      playingFartSound = false
      fadeFart = false
    end
  end
  world:update(dt) --this puts the world into motion
  updateDrawable()
  local cx,cy = cam:pos()
  for i,building in ipairs(drawableBuildings) do
    building:prepare(leftBatches[i], rightBatches[i], frontBatches[i],backBatches[i])
    building:update(cx,cy, 150 * 1/cam.scale)
  end
  pedestrians:update(dt,player)
  if countingDown  and countingDown > -1 then
    countDown(dt)
    if countingDown > 0 then
      return
    elseif countingDown < -1 then
      countingDown = false
    end
  end
  player.isFarting = false
  if love.keyboard.isDown("z") then
    if not playingFartSound then
      fartSound:setVolume(1.0)
      fartSound:setPitch(fartsLeft)
      fartSound:play()
      playingFartSound = true
    end
    if player.remainingFarts > 0 then
      player.isFarting = true
      farts:create(world,player)
    else
      fadeFart = true
    end
    player.remainingFarts = math.max(player.remainingFarts - dt,0)
  else
    player.remainingFarts = math.min(player.remainingFarts + (dt*math.max(0.2,fartsLeft)), player.maxFarts)
  end

  farts:update(dt)

  if gameOver then return end
  currentTime = currentTime + dt
  --here we are going to create some keyboard events
  player:update(dt)

  lowestY = math.min(player.body:getY() + world_height,lowestY)
  player.body:setY(math.min(player.body:getY(), lowestY))

  player:updateAnimation()
  applyEffects(dt)
end

function game:draw()
  
  cam:draw(draw_world)

  draw_hud()
end

function game:focus(f)
  gamePaused = not f or gamePaused
end

return game
