local pedestrians = {}
pedestrians.__index = pedestrians

local function new()
  return setmetatable({
      drawable = {},
      pedestrians = {}
    },pedestrians)
end

function pedestrians:add(world,x,y,image,attitude)
  local pedestrian = {}
  local extra = {}
  extra.animation = newAnimation(image, 64, 64, 0.1, 6)
  extra.angle = math.random(0, 7)
  extra.linePosition = {x,y}
  extra.color = {math.random(200,220),math.random(200,220),math.random(200,220)}
  extra.scale = math.random(4,7)/10
  extra.idx = love.timer.getMicroTime()
  pedestrian.body = love.physics.newBody(world, math.max(x,20), y, "dynamic") --place the body in the center of the world and make it dynamic, so it can move around
  
  if attitude > 0 then
    pedestrian.body:setSleepingAllowed(false)
    pedestrian.body:applyLinearImpulse(math.random(-100,100), math.random(-60,60))
  else
    pedestrian.body:setSleepingAllowed(true)
    pedestrian.body:setAwake(false)
  end
  pedestrian.shape = love.physics.newCircleShape( math.ceil(extra.scale * 28) ) --the player's shape has a radius of 20
  pedestrian.fixture = love.physics.newFixture(pedestrian.body, pedestrian.shape, 13) -- Attach fixture to body and give it a density of 1.
  pedestrian.fixture:setRestitution(0.1) --let the player bounce
  pedestrian.fixture:setCategory(1)
  
  pedestrian.fixture:setUserData(extra)
  table.insert(self.pedestrians, pedestrian)
  pedestrian = nil
  extra = nil
end

function pedestrians:update(dt,player)
  self.drawable = nil
  local drawable = {}
  for i=1, #self.pedestrians do
    if not self.pedestrians[i] then break end
    local x,y = self.pedestrians[i].body:getPosition()
    if math.abs(y - player.body:getY()) < world_height and math.abs(x - player.body:getX()) < world_width then
      self.pedestrians[i].body:setAwake(true)

      local vx,vy = self.pedestrians[i].body:getLinearVelocity()
      local lineX,lineY = unpack(self.pedestrians[i].fixture:getUserData().linePosition)
      local dx = (lineX - x) * 3
      local dy = (lineY - y) * 3
      local mag = math.abs(vx) + math.abs(vy)
      local dMag = math.abs(dx)
      if mag < 15 then
        self.pedestrians[i].fixture:getUserData().animation:reset()
      else
        self.pedestrians[i].fixture:getUserData().animation:setSpeed(mag/200)
      end
      if dMag > 30 then
        self.pedestrians[i].fixture:getUserData().angle = math.atan2(vx, -vy)
      end
      self.pedestrians[i].fixture:getUserData().animation:update(dt)
      self.pedestrians[i].body:applyForce(dx, dy)
      table.insert(drawable, self.pedestrians[i])
    else
      -- self.pedestrians[i].body:setAwake(false)
      if y <= 0 or player.body:getY() + (world_height*2) < y then
        local culled = table.remove(self.pedestrians, i)
        
        culled.fixture:setUserData(nil)
        culled.fixture:destroy()
        culled.fixture = nil
        culled.shape = nil
        culled = nil
      end
    end
  end
  self.drawable = drawable
  drawable = nil
  collectgarbage('collect')
end

function pedestrians:draw()
  for i=1, #self.drawable do
    if self.drawable[i].body then
      love.graphics.setColor(self.drawable[i].fixture:getUserData().color)
      self.drawable[i].fixture:getUserData().animation:draw(self.drawable[i].body:getX(), self.drawable[i].body:getY(), self.drawable[i].fixture:getUserData().angle, 1,1, 32, 32) --pedestrians[i].scale, 128 * pedestrians[i].scale, 128 * pedestrians[i].scale)
    end
  end
end

-- the module
return setmetatable({new = new},
  {__call = function(_, ...) return new(...) end})