local player = {}
player.__index = player

local maxSpeed = 550

local function new(world,x,y,legsAnimation,bodyAnimation,fartFrame)
  local body,shape,fixture
  body = love.physics.newBody(world, x, y, "dynamic")
  body:setLinearDamping(1)
  shape = love.physics.newCircleShape( 7 )
  fixture = love.physics.newFixture(body, shape, 2)
  fixture:setRestitution(0.1)
  fixture:setCategory(3)
  fixture:setMask(2)

  return setmetatable({
    activeEffects = {},
    angle = 0,
    speed = 40,
    scale = 1,
    fartiness = 0.3,
    body = body,
    fixture = fixture,
    maxFarts = 1,
    remainingFarts = 1,
    legsAnimation = legsAnimation,
    bodyAnimation = bodyAnimation,
    fartFrame = fartFrame
  },player)
end

function player:update(dt)

  if love.keyboard.isDown("right") then
    self.body:applyForce(self.speed, 0)
  end
  if love.keyboard.isDown("left") then
    self.body:applyForce(-self.speed, 0)
  end
  if love.keyboard.isDown("up") then
    self.body:applyForce(0,-self.speed)
  end
  if love.keyboard.isDown("down") then
    self.body:applyForce(0,self.speed)
  end
  if love.keyboard.isDown('x') then
    self.body:applyForce(-vx,-vy)
  end

  self.legsAnimation:update(dt)
  self.bodyAnimation:update(dt)
end

function player:renorm()
  self.fartiness = 0.3
  self.speed = 40
  self.scale = 1
  self.body:resetMassData()
end

function player:updateAnimation()
  local x,y = self.body:getLinearVelocity()
  x,y = math.max(math.min(x, maxSpeed),-maxSpeed),math.max(math.min(y, maxSpeed),-maxSpeed)
  self.body:setLinearVelocity(x,y)
  if math.abs(x + y) > 5 then
    self.angle = math.atan2(x,-y)
  end
  
  local mag = math.abs(x) + math.abs(y)
  if mag < 10 then
    self.legsAnimation:reset()
    self.bodyAnimation:reset()
  else
    self.legsAnimation:setSpeed(math.min(mag/100,1.5))
    self.bodyAnimation:setSpeed(math.min(mag/100,1.5))
  end
end

function player:draw()
  self.legsAnimation:draw(self.body:getX(), self.body:getY(), self.angle,self.scale,self.scale,32,32)
  if self.isFarting then
    love.graphics.draw(self.fartFrame, self.body:getX(), self.body:getY(), self.angle + math.pi,self.scale,self.scale,32,32)
  else
    self.bodyAnimation:draw(self.body:getX(), self.body:getY(), self.angle,self.scale,self.scale,32,32)
  end
end
-- the module
return setmetatable({new = new},
  {__call = function(_, ...) return new(...) end})